@@ -77,7 +77,7 @@ sub new {
       LINKARGS => [],  # Linker args
       NATIVECAML => 1, # this causes the native code boxer to be used
       RELEASELIB => 0, # if true, use the release runtime library (if any)
-      IDASHI => 1,     # if true, pass "-I-" to gcc's preprocessor
+      # IDASHI => 1,     # if true, pass "-I-" to gcc's preprocessor
       IDASHDOT => 1,   # if true, pass "-I." to gcc's preprocessor
       VERBOSE => 0,    # when true, print extra detail
       TRACE_COMMANDS => 1, # when true, echo commands being run
@@ -201,7 +201,12 @@ sub collectOneArgument {
     if($self->compilerArgument($self->{OPTIONS}, $arg, $pargs)) { return 1; }
 
     if($arg eq "--help" || $arg eq "-help") {
-        $self->printHelp(); exit 1;
+        $self->printVersion(); 
+        $self->printHelp(); 
+        exit 1;
+    }
+    if($arg eq "--version" || $arg eq "-version") {
+        $self->printVersion(); exit 0;
     }
     if($arg eq "--verbose") {
         $self->{VERBOSE} = 1; return 1;
@@ -225,11 +230,19 @@ sub collectOneArgument {
         $self->{TRUEOBJ} = 1;
         return 1;
     }
+    # zf: force curing when linking to a lib
+    if ($arg eq '--truelib') {
+	$self->{TRUELIB} = 1;
+	return 1;
+    }
     if($arg eq '--keepmerged') {
         $self->{KEEPMERGED} = 1;
         return 1;
     }
-
+    if($arg eq '--stdoutpp') {
+        $self->{STDOUTPP} = 1;
+        return 1;
+    }
     if($arg =~ m|--save-temps=(.+)$|) {
         if(! -d $1) {
             die "Cannot find directory $1";
@@ -256,9 +269,9 @@ sub collectOneArgument {
     if($arg eq "--bytecode") {
         $self->{NATIVECAML} = 0; return 1;
     }
-    if($arg eq "--no-idashi") {
-        $self->{IDASHI} = 0; return 1;
-    }
+#     if($arg eq "--no-idashi") {
+#         $self->{IDASHI} = 0; return 1;
+#     }
     if($arg eq "--no-idashdot") {
         $self->{IDASHDOT} = 0; return 1;
     }
@@ -328,11 +341,15 @@ sub collectOneArgument {
 }
 
 
+sub printVersion {
+    system ($CilCompiler::compiler, '--version'); 
+}
 
 sub printHelp {
     my($self) = @_;
     $self->usage();
     print <<EOF;
+
 Options:
   --mode=xxx   What tool to emulate:
                 GNUCC   - GNU gcc
@@ -353,15 +370,16 @@ Options:
   --keepmerged  Save the merged file. Only useful if --nomerge is not given.
   --trueobj          Do not write preprocessed sources in .obj/.o files but
                      create some other files (e.g. foo.o_saved.c).
- 
+  --truelib          When linking to a library (with -r or -i), output real
+                     object files instead of preprocessed sources. This only
+		     works for GCC right now.
   --leavealone=xxx   Leave alone files whose base name is xxx. This means
                      they are not merged and not processed with CIL.
   --includedir=xxx   Adds a new include directory to replace existing ones
-  --stages           Show the processing stages
   --bytecode         Invoke the bytecode (as opposed to native code) system
-  --no-idashi        Do not use '-I-' with the gcc preprocessor.
 
 EOF
+#   --no-idashi        Do not use '-I-' with the gcc preprocessor.
     $self->helpMessage();
 }
 
@@ -413,7 +431,8 @@ sub linktolib {
     if($self->{VERBOSE}) { print STDERR "Linking into library $dest\n"; }
 
     # Now collect the files to be merged
-    my ($tomerge, $trueobjs) = $self->separateTrueObjects($psrcs);
+    my ($tomerge, $trueobjs, $ccargs) = 
+        $self->separateTrueObjects($psrcs, $ccargs);
 
     if($self->{SEPARATE} || @{$tomerge} == 0) {
         # Not merging. Regular linking.
@@ -467,7 +486,12 @@ sub linktolib {
     if(@{$tomerge} > 20) {
         my $extraFile = "___extra_files";
         open(TOMERGE, ">$extraFile") || die $!;
-        foreach my $fl (@{$tomerge}) {
+        #FRANJO added the following on February 15th, 2005
+        #REASON: extrafiles was TempFIle=HASH(0x12345678) 
+        # instead of actual filename
+        my @normalized = @{$tomerge} ;
+        $_ = (ref $_ ? $_->filename : $_) foreach @normalized;
+        foreach my $fl (@normalized) {
             print TOMERGE "$fl\n";
         }
         close(TOMERGE);
@@ -567,9 +591,10 @@ sub preprocess_before_cil {
             # for apache.. more investigation is needed
             # update: now when I try it, apache works without -I- also.. but
             # I'll make this into a switchable flag anyway
-            if ($self->{IDASHI}) {
-                unshift @args, "-I-";
-            }
+            # matth: this breaks other tests.  Let's try without.
+#             if ($self->{IDASHI}) {
+#                 unshift @args, "-I-";
+#             }
         }
     }
 
@@ -642,18 +667,16 @@ sub compile {
         $mtime = $mtime_1;
         $outfile = $dest->{filename} . $Cilly::savedSourceExt;
     }
-    # sm: made the following output unconditional following the principle
-    # that by default you should be able to see every file getting written
-    # during a build (otherwise you don't know who to ask to be --verbose)
-    # update: and then someone reverted it to conditional... why?
+    my $srcname = ref $src ? $src->filename : $src; 
     if($self->{VERBOSE}) { 
-        print STDERR "Saving source $src->{filename} into $outfile\n";
+        print STDERR "Saving source $srcname into $outfile\n";
     }
     open(OUT, ">$outfile") || die "Cannot create $outfile";
-    my $toprintsrc = $src->{filename}; $toprintsrc =~ s|\\|/|g;
-    print OUT "#pragma merger($mtime, \"$toprintsrc\", \"" . 
-        join(' ', @{$ccargs}), "\")\n";
-    open(IN, '<', $src->{filename}) || die "Cannot read $src->{filename}";
+    my $toprintsrc = $srcname; 
+    $toprintsrc =~ s|\\|/|g;
+    print OUT "#pragma merger($mtime,\"$toprintsrc\",\"" . 
+              join(',', @{$ccargs}), "\")\n";
+    open(IN, '<', $srcname) || die "Cannot read $srcname";
     while(<IN>) {
         print OUT $_;
     }
@@ -679,7 +702,7 @@ sub makeOutArguments {
 sub straight_compile {
     my ($self, $src, $dest, $ppargs, $ccargs) = @_;
     if($self->{VERBOSE}) { 
-        print STDERR 'Compiling ', $src->filename, ' into ', 
+        print STDERR 'Compiling ', ref $src ? $src->filename : $src, ' into ', 
         $dest->filename, "\n"; 
     }
     my @dest = 
@@ -764,12 +787,14 @@ sub expandLibraries {
 # actually sources to be merged, and the true object files
 #
 sub separateTrueObjects {
-    my ($self, $psrcs) = @_;
+    my ($self, $psrcs, $ccargs) = @_;
 
     my @sources = @{$psrcs};
 #    print "Sources are @sources\n";
     my @tomerge = ();
     my @othersources = ();
+
+    my @ccmerged = @{$ccargs};
     foreach my $src (@sources) {
         my ($combsrc, $combsrcname, $mtime);
 	my $srcname = ref $src ? $src->filename : $src;
@@ -793,8 +818,22 @@ sub separateTrueObjects {
         open(IN, "<$combsrcname") || die "Cannot read $combsrcname";
         my $fstline = <IN>;
         close(IN);
-        if($fstline =~ m|\#pragma merger\((\d+)| || $fstline =~ m|CIL|) {
-            if($1 == $mtime) { # It is ours
+        if($fstline =~ m|CIL|) {
+            goto ToMerge;
+        } 
+        if($fstline =~ m|^\#pragma merger\((\d+),\".*\",\"(.*)\"\)$|) {
+            my $mymtime = $1;
+            # Get the CC flags
+            my @thisccargs = split(/,/, $2);
+            foreach my $arg (@thisccargs) {
+                # print "Looking at $arg\n  ccmerged=@ccmerged\n";
+                if(! grep(/$arg/, @ccmerged)) {
+                    # print " adding it\n";
+                    push @ccmerged, $arg
+                }
+            }
+          ToMerge:
+            if($mymtime == $mtime) { # It is ours
                 # See if we have this already
                 if(! grep { $_ eq $srcname } @tomerge) { # It is ours
                     push @tomerge, $combsrc; 
@@ -815,7 +854,13 @@ sub separateTrueObjects {
         }
         push @othersources, $combsrc;
     }
-    return (\@tomerge, \@othersources);
+    # If we are merging, turn off "warnings are errors" flag
+    if(grep(/$self->{WARNISERROR}/, @ccmerged)) {
+        @ccmerged = grep(!/$self->{WARNISERROR}/, @ccmerged);
+        print STDERR "Turning off warn-is-error flag $self->{WARNISERROR}\n";
+    }
+
+    return (\@tomerge, \@othersources, \@ccmerged);
 }
 
 
@@ -844,13 +889,15 @@ sub link {
     
     # Now collect the files to be merged
 
-    my ($tomerge, $trueobjs) = $self->separateTrueObjects($psrcs);
+    my ($tomerge, $trueobjs, $ccargs) = 
+        $self->separateTrueObjects($psrcs, $ccargs);
 
     if($self->{VERBOSE}) {
         print STDERR "Will merge the following: ", 
                          join(' ', @{$tomerge}), "\n";
         print STDERR "Will just link the genuine object files: ", 
                          join(' ', @{$trueobjs}), "\n";
+        print STDERR "After merge compile flags: @{$ccargs}\n";
     }
     # Check the modification times and see if we can just use the combined
     # file instead of merging all over again
@@ -878,6 +925,13 @@ sub link {
     unshift @{$trueobjs}, $mergedobj;
 
     # And finally link
+    # zf: hack for linking linux stuff
+    if ($self->{TRUELIB}) {
+	my @cmd = (@{$self->{LDLIB}}, ($dest),
+		   @{$ppargs}, @{$ccargs}, @{$trueobjs}, @{$ldargs});
+	return $self->runShell(@cmd);
+    }
+
     # sm: hack: made this conditional for dsw
     if (!defined($ENV{CILLY_DONT_LINK_AFTER_MERGE})) {
       $self->link_after_cil($trueobjs, $dest, $ppargs, $ccargs, $ldargs);
@@ -913,7 +967,8 @@ sub applyCil {
         my $extraFile = "___extra_files";
         open(TOMERGE, ">$extraFile") || die $!;
         foreach my $fl (@srcs) {
-            print TOMERGE "$fl\n";
+            my $fname =  ref $fl ? $fl->filename : $fl;
+            print TOMERGE "$fname\n";
         }
         close(TOMERGE);
         push @cmd, '--extrafiles', $extraFile;
@@ -959,7 +1014,9 @@ sub applyCilAndCompile {
 # Linking after CIL
 sub link_after_cil {
     my ($self, $psrcs, $dest, $ppargs, $ccargs, $ldargs) = @_;
-    return $self->straight_link($psrcs, $dest, $ppargs, $ccargs, $ldargs);
+    if (!defined($ENV{CILLY_DONT_COMPILE_AFTER_MERGE})) {
+      return $self->straight_link($psrcs, $dest, $ppargs, $ccargs, $ldargs);
+    }
 }
 
 # See if we must merge this one
@@ -1007,8 +1064,9 @@ sub doit {
         # object file is a disguised source
         my $turnOffMerging = 0;
         if(@{$self->{OFILES}}) {
-            my ($tomerge, $trueobjs) = 
-                $self->separateTrueObjects($self->{OFILES});
+            my ($tomerge, $trueobjs, $mergedccargs) = 
+                $self->separateTrueObjects($self->{OFILES}, $self->{CCARGS});
+            $self->{CCARGS} = $mergedccargs;
             $turnOffMerging = (@{$tomerge} == 0);
         } else {
             $turnOffMerging = 1;
@@ -1049,11 +1107,27 @@ sub doit {
 
     # See if we must create a library only
     if($self->{OPERATION} eq "TOLIB") {
-        $out = $self->linkOutputFile(@tolink);
-        $self->linktolib(\@tolink,  $out, 
-                         $self->{PPARGS}, $self->{CCARGS}, 
-                         $self->{LINKARGS});
-        return;
+	if (!$self->{TRUELIB}) {
+	    # zf: Creating a library containing merged source
+	    $out = $self->linkOutputFile(@tolink);
+	    $self->linktolib(\@tolink,  $out, 
+			     $self->{PPARGS}, $self->{CCARGS}, 
+			     $self->{LINKARGS});
+	    return;
+	} else {
+	    # zf: Linking to a true library. Do real curing.
+	    # Only difference from TOEXE is that we use "partial linking" of the 
+	    # underlying linker
+	    if ($self->{VERBOSE}) {
+	        print STDERR "Linking to a true library!";
+	    }
+	    push @{$self->{CCARGS}}, "-r";
+	    $out = $self->linkOutputFile(@tolink);
+	    $self->link(\@tolink,  $out, 
+			$self->{PPARGS}, $self->{CCARGS}, $self->{LINKARGS});
+	    return;
+	}
+
     }
 
     # Now link all of the files into an executable
@@ -1063,6 +1137,7 @@ sub doit {
                     $self->{PPARGS}, $self->{CCARGS}, $self->{LINKARGS});
         return;
     }
+
     die "I don't understand OPERATION:$self->{OPERATION}\n";
 }
 
@@ -1294,6 +1369,7 @@ sub new {
       EXEEXT => ".exe",  # Executable extension (with the .)
       OUTOBJ => "/Fo",
       OUTEXE => "/Fe",
+      WARNISERROR => "/WX",
       FORCECSOURCE => ['/Tc'],
       LINEPATTERN => "^#line\\s+(\\d+)\\s+\"(.+)\"",
 
@@ -1380,11 +1456,17 @@ sub msvc_preprocess {
     if($sext eq ".cpp") {
         push @cmd, "/Tc";
     }
-    @cmd = ('cl', '/nologo', '/P', '/D_MSVC', @cmd);
-    $res = $self->runShell(@cmd, $srcname);
     # MSVC cannot be told where to put the output. But we know that it
     # puts it in the current directory
     my $msvcout = "./$sbase.i";
+    if($self->{STDOUTPP}) {
+        @cmd = ('cmd', '/c', 'cl', '/nologo', '/E', ">$msvcout", '/D_MSVC', 
+                @cmd);
+        
+    } else { 
+        @cmd = ('cl', '/nologo', '/P', '/D_MSVC', @cmd);
+    }
+    $res = $self->runShell(@cmd, $srcname);
     # Check file equivalence by making sure that all elements of the stat
     # structure are the same, except for the access time.
     my @st1 = stat $msvcout; $st1[8] = 0;
@@ -1532,6 +1614,7 @@ sub new {
       EXEEXT => ".exe",  # Executable extension (with the .)
       OUTOBJ => $msvc->{OUTOBJ},
       OUTEXE => "-out:", # Keep this form because build.exe looks for it
+      WARNISERROR => "/WX",
       LINEPATTERN => "", 
       FORCECSOURCE => $msvc->{FORCECSOURCE},
 
@@ -1794,6 +1877,7 @@ sub new {
       OUTOBJ => '-o',
       OUTEXE => '-o',
       OUTCPP => '-o',
+      WARNISERROR => "-Werror",
       FORCECSOURCE => [],
       LINEPATTERN => "^#\\s+(\\d+)\\s+\"(.+)\"",
       
@@ -1806,11 +1890,13 @@ sub new {
             "-E"   => { RUN => sub { $stub->{OPERATION} = "TOI"; }},
 	    "-pipe\$" => { TYPE => 'ALLARGS' },
             "-[DIU]" => { ONEMORE => 1, TYPE => "PREPROC" },
+            "-isystem" => { ONEMORE => 1, TYPE => "PREPROC" },
             '-undef$' => { TYPE => 'PREPROC' },
             '-w$' => { TYPE => 'PREPROC' },
 	    '-M$' => { TYPE => 'SPECIAL' },
 	    '-MM$' => { TYPE => 'SPECIAL' },
 	    '-MF$' => { TYPE => 'EARLY_PREPROC', ONEMORE => 1 },
+	    '-C$' =>  { TYPE => 'EARLY_PREPROC'}, # zra
 	    '-MG$' => { TYPE => 'EARLY_PREPROC' },
 	    '-MP$' => { TYPE => 'EARLY_PREPROC' },
 	    '-MT$' => { TYPE => 'EARLY_PREPROC', ONEMORE => 1 },
