@@ -68,7 +68,7 @@ let warnContext fmt =
 let checkAttributes (attrs: attribute list) : unit = 
   let rec loop lastname = function
       [] -> ()
-    | (Attr(an, _) as a) :: resta -> 
+    | Attr(an, _) :: resta -> 
         if an < lastname then
           ignore (warn "Attributes not sorted");
         loop an resta
@@ -453,6 +453,7 @@ and checkExp (isconst: bool) (e: exp) : typ =
       | Const(CStr s) -> charPtrType
       | Const(CWStr s) -> TPtr(!wcharType,[])
       | Const(CReal (_, fk, _)) -> TFloat(fk, [])
+      | Const(CEnum (_, _, ei)) -> TEnum(ei, [])
       | Lval(lv) -> 
           if isconst then
             ignore (warn "Lval in constant");
